@@ -31,57 +31,109 @@ WiFi Manager for making ESP32 SPI as WiFi much easier
 
 # pylint: disable=no-name-in-module
 
+from time import sleep
+from micropython import const
 from adafruit_esp32spi import adafruit_esp32spi
-import adafruit_esp32spi.adafruit_esp32spi_requests as requests
+import adafruit_esp32spi.adafruit_esp32spi_socket as socket
+import adafruit_requests as requests
+
 
 class ESPSPI_WiFiManager:
     """
     A class to help manage the Wifi connection
     """
-    def __init__(self, esp, secrets, status_pixel=None, attempts=2):
+
+    NORMAL = const(1)
+    ENTERPRISE = const(2)
+
+    # pylint: disable=too-many-arguments
+    def __init__(
+        self,
+        esp,
+        secrets,
+        status_pixel=None,
+        attempts=2,
+        connection_type=NORMAL,
+        debug=False,
+    ):
         """
         :param ESP_SPIcontrol esp: The ESP object we are using
         :param dict secrets: The WiFi and Adafruit IO secrets dict (See examples)
-        :param status_pixel: (Optional) The pixel device - A NeoPixel or DotStar (default=None)
-        :type status_pixel: NeoPixel or DotStar
+        :param status_pixel: (Optional) The pixel device - A NeoPixel, DotStar,
+            or RGB LED (default=None)
+        :type status_pixel: NeoPixel, DotStar, or RGB LED
         :param int attempts: (Optional) Failed attempts before resetting the ESP32 (default=2)
+        :param const connection_type: (Optional) Type of WiFi connection: NORMAL or ENTERPRISE
         """
         # Read the settings
-        self._esp = esp
-        self.debug = False
-        self.ssid = secrets['ssid']
-        self.password = secrets['password']
+        self.esp = esp
+        self.debug = debug
+        self.ssid = secrets["ssid"]
+        self.password = secrets.get("password", None)
         self.attempts = attempts
-        requests.set_interface(self._esp)
+        self._connection_type = connection_type
+        requests.set_socket(socket, esp)
         self.statuspix = status_pixel
         self.pixel_status(0)
 
+        # Check for WPA2 Enterprise keys in the secrets dictionary and load them if they exist
+        if secrets.get("ent_ssid"):
+            self.ent_ssid = secrets["ent_ssid"]
+        else:
+            self.ent_ssid = secrets["ssid"]
+        if secrets.get("ent_ident"):
+            self.ent_ident = secrets["ent_ident"]
+        else:
+            self.ent_ident = ""
+        if secrets.get("ent_user"):
+            self.ent_user = secrets["ent_user"]
+        if secrets.get("ent_password"):
+            self.ent_password = secrets["ent_password"]
+
+    # pylint: enable=too-many-arguments
+
     def reset(self):
         """
         Perform a hard reset on the ESP32
         """
         if self.debug:
             print("Resetting ESP32")
-        self._esp.reset()
+        self.esp.reset()
 
     def connect(self):
         """
         Attempt to connect to WiFi using the current settings
         """
         if self.debug:
-            if self._esp.status == adafruit_esp32spi.WL_IDLE_STATUS:
+            if self.esp.status == adafruit_esp32spi.WL_IDLE_STATUS:
                 print("ESP32 found and in idle mode")
-            print("Firmware vers.", self._esp.firmware_version)
-            print("MAC addr:", [hex(i) for i in self._esp.MAC_address])
-            for access_pt in self._esp.scan_networks():
-                print("\t%s\t\tRSSI: %d" % (str(access_pt['ssid'], 'utf-8'), access_pt['rssi']))
+            print("Firmware vers.", self.esp.firmware_version)
+            print("MAC addr:", [hex(i) for i in self.esp.MAC_address])
+            for access_pt in self.esp.scan_networks():
+                print(
+                    "\t%s\t\tRSSI: %d"
+                    % (str(access_pt["ssid"], "utf-8"), access_pt["rssi"])
+                )
+        if self._connection_type == ESPSPI_WiFiManager.NORMAL:
+            self.connect_normal()
+        elif self._connection_type == ESPSPI_WiFiManager.ENTERPRISE:
+            self.connect_enterprise()
+        else:
+            raise TypeError("Invalid WiFi connection type specified")
+
+    def connect_normal(self):
+        """
+        Attempt a regular style WiFi connection
+        """
         failure_count = 0
-        while not self._esp.is_connected:
+        while not self.esp.is_connected:
             try:
                 if self.debug:
                     print("Connecting to AP...")
                 self.pixel_status((100, 0, 0))
-                self._esp.connect_AP(bytes(self.ssid, 'utf-8'), bytes(self.password, 'utf-8'))
+                self.esp.connect_AP(
+                    bytes(self.ssid, "utf-8"), bytes(self.password, "utf-8")
+                )
                 failure_count = 0
                 self.pixel_status((0, 100, 0))
             except (ValueError, RuntimeError) as error:
@@ -92,6 +144,64 @@ class ESPSPI_WiFiManager:
                     self.reset()
                 continue
 
+    def create_ap(self):
+        """
+        Attempt to initialize in Access Point (AP) mode.
+        Uses SSID and optional passphrase from the current settings
+        Other WiFi devices will be able to connect to the created Access Point
+        """
+        failure_count = 0
+        while not self.esp.ap_listening:
+            try:
+                if self.debug:
+                    print("Waiting for AP to be initialized...")
+                self.pixel_status((100, 0, 0))
+                if self.password:
+                    self.esp.create_AP(
+                        bytes(self.ssid, "utf-8"), bytes(self.password, "utf-8")
+                    )
+                else:
+                    self.esp.create_AP(bytes(self.ssid, "utf-8"), None)
+                failure_count = 0
+                self.pixel_status((0, 100, 0))
+            except (ValueError, RuntimeError) as error:
+                print("Failed to create access point\n", error)
+                failure_count += 1
+                if failure_count >= self.attempts:
+                    failure_count = 0
+                    self.reset()
+                continue
+        print("Access Point created! Connect to ssid:\n {}".format(self.ssid))
+
+    def connect_enterprise(self):
+        """
+        Attempt an enterprise style WiFi connection
+        """
+        failure_count = 0
+        self.esp.wifi_set_network(bytes(self.ent_ssid, "utf-8"))
+        self.esp.wifi_set_entidentity(bytes(self.ent_ident, "utf-8"))
+        self.esp.wifi_set_entusername(bytes(self.ent_user, "utf-8"))
+        self.esp.wifi_set_entpassword(bytes(self.ent_password, "utf-8"))
+        self.esp.wifi_set_entenable()
+        while not self.esp.is_connected:
+            try:
+                if self.debug:
+                    print(
+                        "Waiting for the ESP32 to connect to the WPA2 Enterprise AP..."
+                    )
+                self.pixel_status((100, 0, 0))
+                sleep(1)
+                failure_count = 0
+                self.pixel_status((0, 100, 0))
+                sleep(1)
+            except (ValueError, RuntimeError) as error:
+                print("Failed to connect, retrying\n", error)
+                failure_count += 1
+                if failure_count >= self.attempts:
+                    failure_count = 0
+                    self.reset()
+                continue
+
     def get(self, url, **kw):
         """
         Pass the Get request to requests and update status LED
@@ -104,7 +214,7 @@ class ESPSPI_WiFiManager:
         :return: The response from the request
         :rtype: Response
         """
-        if not self._esp.is_connected:
+        if not self.esp.is_connected:
             self.connect()
         self.pixel_status((0, 0, 100))
         return_val = requests.get(url, **kw)
@@ -123,7 +233,7 @@ class ESPSPI_WiFiManager:
         :return: The response from the request
         :rtype: Response
         """
-        if not self._esp.is_connected:
+        if not self.esp.is_connected:
             self.connect()
         self.pixel_status((0, 0, 100))
         return_val = requests.post(url, **kw)
@@ -141,7 +251,7 @@ class ESPSPI_WiFiManager:
         :return: The response from the request
         :rtype: Response
         """
-        if not self._esp.is_connected:
+        if not self.esp.is_connected:
             self.connect()
         self.pixel_status((0, 0, 100))
         return_val = requests.put(url, **kw)
@@ -160,7 +270,7 @@ class ESPSPI_WiFiManager:
         :return: The response from the request
         :rtype: Response
         """
-        if not self._esp.is_connected:
+        if not self.esp.is_connected:
             self.connect()
         self.pixel_status((0, 0, 100))
         return_val = requests.patch(url, **kw)
@@ -179,7 +289,7 @@ class ESPSPI_WiFiManager:
         :return: The response from the request
         :rtype: Response
         """
-        if not self._esp.is_connected:
+        if not self.esp.is_connected:
             self.connect()
         self.pixel_status((0, 0, 100))
         return_val = requests.delete(url, **kw)
@@ -195,10 +305,10 @@ class ESPSPI_WiFiManager:
         :return: The response time in milliseconds
         :rtype: int
         """
-        if not self._esp.is_connected:
+        if not self.esp.is_connected:
             self.connect()
         self.pixel_status((0, 0, 100))
-        response_time = self._esp.ping(host, ttl=ttl)
+        response_time = self.esp.ping(host, ttl=ttl)
         self.pixel_status(0)
         return response_time
 
@@ -206,26 +316,29 @@ class ESPSPI_WiFiManager:
         """
         Returns a formatted local IP address, update status pixel.
         """
-        if not self._esp.is_connected:
+        if not self.esp.is_connected:
             self.connect()
         self.pixel_status((0, 0, 100))
         self.pixel_status(0)
-        return self._esp.pretty_ip(self._esp.ip_address)
+        return self.esp.pretty_ip(self.esp.ip_address)
 
     def pixel_status(self, value):
         """
-        Change Status NeoPixel if it was defined
+        Change Status Pixel if it was defined
 
         :param value: The value to set the Board's status LED to
         :type value: int or 3-value tuple
         """
         if self.statuspix:
-            self.statuspix.fill(value)
+            if hasattr(self.statuspix, "color"):
+                self.statuspix.color = value
+            else:
+                self.statuspix.fill(value)
 
     def signal_strength(self):
         """
         Returns receiving signal strength indicator in dBm
         """
-        if not self._esp.is_connected:
+        if not self.esp.is_connected:
             self.connect()
-        return self._esp.rssi()
+        return self.esp.rssi()
