@@ -31,6 +31,7 @@ https://github.com/adafruit/Adafruit_Blinka/blob/master/src/digitalio.py
 """
 from micropython import const
 
+
 class Pin:
     """
     Implementation of CircuitPython API Pin Handling
@@ -42,7 +43,8 @@ class Pin:
     NOTE: This class does not currently implement reading digital pins
     or the use of internal pull-up resistors.
     """
-     #pylint: disable=invalid-name
+
+    # pylint: disable=invalid-name
     IN = const(0x00)
     OUT = const(0x01)
     LOW = const(0x00)
@@ -51,17 +53,15 @@ class Pin:
     _mode = IN
     pin_id = None
 
-    ESP32_GPIO_PINS = set([0, 1, 2, 4, 5,
-                           12, 13, 14, 15,
-                           16, 17, 18, 19,
-                           21, 22, 23, 25,
-                           26, 27, 32, 33])
+    ESP32_GPIO_PINS = set(
+        [0, 1, 2, 4, 5, 12, 13, 14, 15, 16, 17, 18, 19, 21, 22, 23, 25, 26, 27, 32, 33]
+    )
 
     def __init__(self, esp_pin, esp):
         if esp_pin in self.ESP32_GPIO_PINS:
             self.pin_id = esp_pin
         else:
-            raise AttributeError("Pin %d is not a valid ESP32 GPIO Pin."%esp_pin)
+            raise AttributeError("Pin %d is not a valid ESP32 GPIO Pin." % esp_pin)
         self._esp = esp
 
     def init(self, mode=IN):
@@ -92,36 +92,46 @@ class Pin:
             else:
                 raise RuntimeError("Invalid value for pin")
         else:
-            raise NotImplementedError("digitalRead not currently implemented in esp32spi")
+            raise NotImplementedError(
+                "digitalRead not currently implemented in esp32spi"
+            )
 
     def __repr__(self):
         return str(self.pin_id)
 
+
 # pylint: disable = too-few-public-methods
-class DriveMode():
+class DriveMode:
     """DriveMode Enum."""
+
     PUSH_PULL = None
     OPEN_DRAIN = None
+
+
 DriveMode.PUSH_PULL = DriveMode()
 DriveMode.OPEN_DRAIN = DriveMode()
 
 
-class Direction():
+class Direction:
     """DriveMode Enum."""
+
     INPUT = None
     OUTPUT = None
+
+
 Direction.INPUT = Direction()
 Direction.OUTPUT = Direction()
 
 
-class DigitalInOut():
+class DigitalInOut:
     """Implementation of DigitalIO module for ESP32SPI.
 
     :param ESP_SPIcontrol esp: The ESP object we are using.
     :param int pin: Valid ESP32 GPIO Pin, predefined in ESP32_GPIO_PINS.
     """
+
     _pin = None
-    #pylint: disable = attribute-defined-outside-init
+    # pylint: disable = attribute-defined-outside-init
     def __init__(self, esp, pin):
         self._esp = esp
         self._pin = Pin(pin, self._esp)
@@ -150,7 +160,9 @@ class DigitalInOut():
         """Sets the pull and then switch to read in digital values.
         :param Pull pull: Pull configuration for the input.
         """
-        raise NotImplementedError("Digital reads are not currently supported in ESP32SPI.")
+        raise NotImplementedError(
+            "Digital reads are not currently supported in ESP32SPI."
+        )
 
     @property
     def direction(self):
@@ -175,7 +187,7 @@ class DigitalInOut():
     @property
     def value(self):
         """Returns the digital logic level value of the pin."""
-        return self._pin.value() is 1
+        return self._pin.value() == 1
 
     @value.setter
     def value(self, val):
@@ -204,6 +216,8 @@ class DigitalInOut():
         """
         self.__drive_mode = mode
         if mode is DriveMode.OPEN_DRAIN:
-            raise NotImplementedError('Drive mode %s not implemented in ESP32SPI.'%mode)
-        elif mode is DriveMode.PUSH_PULL:
+            raise NotImplementedError(
+                "Drive mode %s not implemented in ESP32SPI." % mode
+            )
+        if mode is DriveMode.PUSH_PULL:
             self._pin.init(mode=Pin.OUT)
